from multiprocessing.pool import ThreadPool as Pool

from .utils import test_call, SUBPROCESS_TIMEOUT


def call_concurrently(concurrency, function, **kwargs):
    """
    Make identical concurrent calls to a function.

    Args:
        concurrency (int) - how many calls to make in parallel
        function (Union[function, str]) - the function to call, or
            the 'dotted module.path.to:function' as a string (NOTE
            colon separates the name to import)
            NOTE:
            when `f` is a decorated function where decorator returns a new
            object and functools.wraps was not used (and maybe some other
            cases too) we need to be able to tell our subprocess how to
            import `f`... in this case using string path is mandatory (as
            we cannot introspect it)
        **kwargs - kwargs to pass to `function`

    Returns:
        List[Any] - return values from each run `function`
            (results are returned in order that tasks were enqueued)

    NOTE:
        `kwargs` must be pickleable
    """
    return make_concurrent_calls(
        *[(function, kwargs) for i in range(concurrency)]
    )


def make_concurrent_calls(*calls):
    """
    If you need to make multiple concurrent calls, potentially to
    different functions, or with different kwargs each time.

    Args:
        *calls (Iterable[Union[function, str], dict]) - list of
            (func or func path, kwargs) tuples to call concurrently

    Returns:
        List[Any] - return values from each call in `calls`
            (results are returned in same order as supplied)
    """
    pool = Pool(len(calls))
    results = []
    for func, kwargs in calls:
        results.append(
            pool.apply_async(test_call, args=(func,), kwds=kwargs)
        )
    pool.close()
    pool.join()
    # add a bit of extra timeout to allow process terminate cleanup to run
    # (because we also have an inner timeout on our ProcessManager thread join)
    return [result.get(timeout=SUBPROCESS_TIMEOUT + 2) for result in results]



import pickle
from base64 import b64decode, b64encode

import six


__all__ = ('dumps', 'loads')


"""
we base64 encode to ensure only ascii chars are present
http://bugs.python.org/issue2980
"""


def dumps(obj):
    encoded = b64encode(pickle.dumps(obj, protocol=0))
    if six.PY3:
        # above returns bytes, we need to be able to use as str
        encoded = encoded.decode(encoding='ascii')
    return encoded


def loads(val):
    if six.PY3 and not isinstance(val, six.binary_type):
        # Python 3.2 requires a bytestring, later Py3s don't care
        val = bytes(val, encoding='ascii')
    return pickle.loads(b64decode(val))



from __future__ import print_function
import os
import pickle
import subprocess
import sys
import threading
from contextlib import contextmanager

from django.conf import settings
from django.core.management import call_command
import six

from . import b64pickle


SUBPROCESS_TIMEOUT = int(os.environ.get('DJANGO_CONCURRENT_TESTS_TIMEOUT', '30'))


class UnpickleableError(Exception):
    pass


class TerminatedProcessError(Exception):
    pass


class ProcessManager(object):
    def __init__(self, cmd):
        self.cmd = cmd
        self.process = None
        self.stdout = None
        self.stderr = None
        self.terminated = False

    def run(self, timeout):
        def target():
            self.process = subprocess.Popen(
                self.cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            self.stdout, self.stderr = self.process.communicate()

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(timeout)
        if thread.is_alive():
            # we reached the timeout deadline with process still running
            self.process.terminate()
            self.terminated = True
            thread.join()

        print(self.stderr)
        return self.stdout


def test_call(f, **kwargs):
    """
    Args:
        f (Union[function, str]) - the function to call, or
            the 'dotted module.path.to:function' as a string (NOTE
            colon separates the name to import)
        **kwargs - kwargs to pass to `function`

    Returns:
        Any - either:
            <return value> OR <exception raised>

    NOTE:
        `kwargs` must be pickleable
        <return value> of `function` must be pickleable
    """
    # wrap everything in a catch-all except to avoid hanging the subprocess
    try:
        serialized_kwargs = b64pickle.dumps(kwargs)

        if isinstance(f, six.string_types):
            function_path = f
        else:
            function_path = '{module}:{name}'.format(
                module=f.__module__,
                name=f.__name__,
            )

        print('Calling {f} in subprocess'.format(f=function_path))

        if not os.environ.get('CONCURRENT_TESTS_NO_SUBPROCESS'):
            cmd = [
                getattr(settings, 'MANAGE_PY_PATH', 'manage.py'),
                'concurrent_call_wrapper',
                function_path,
                '--kwargs=%s' % serialized_kwargs,
            ]
            manager = ProcessManager(cmd)
            result = manager.run(timeout=SUBPROCESS_TIMEOUT)
            if manager.terminated:
                raise TerminatedProcessError(result)
        else:
            # TODO: collect stdout
            result = call_command(
                'concurrent_call_wrapper',
                function_path,
                kwargs=serialized_kwargs,
            )
    except Exception as e:
        try:
            pickle.dumps(e)
        except Exception:
            return UnpickleableError(repr(e))
        else:
            return e
    return b64pickle.loads(result) if result else None


@contextmanager
def redirect_stdout(to):
    original = sys.stdout
    sys.stdout = to
    yield
    sys.stdout = original
